@@ -1,18 +1,18 @@
 /*
  * This file is a part of Java Bindings for Librhash
- * Copyright (c) 2011-2012, 2014, Sergey Basalaev <sbasalaev@gmail.com>
- * Librhash is (c) 2011-2014, Aleksey Kravchenko <rhash.admin@gmail.com>
- * 
- * Permission is hereby granted, free of charge,  to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to deal
- * in the Software without restriction,  including without limitation the rights
- * to  use,  copy,  modify,  merge, publish, distribute, sublicense, and/or sell
- * copies  of  the Software,  and  to permit  persons  to whom  the Software  is
- * furnished to do so.
- * 
- * This library  is distributed  in the hope that it will be useful, but WITHOUT
- * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
- * FOR A PARTICULAR PURPOSE. Use it at your own risk!
+ *
+ * Copyright (c) 2011, Sergey Basalaev <sbasalaev@gmail.com>
+ *
+ * Permission to use, copy, modify, and/or distribute this software for any
+ * purpose with or without fee is hereby granted.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
+ * REGARD TO THIS SOFTWARE  INCLUDING ALL IMPLIED WARRANTIES OF  MERCHANTABILITY
+ * AND FITNESS.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
+ * INDIRECT,  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER RESULTING FROM
+ * LOSS OF USE,  DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT,  NEGLIGENCE
+ * OR OTHER TORTIOUS ACTION,  ARISING OUT OF  OR IN CONNECTION  WITH THE USE  OR
+ * PERFORMANCE OF THIS SOFTWARE.
  */
 
 package org.sf.rhash;
@@ -74,13 +74,13 @@ public enum HashType {
 	/** SHA3-384 hash. */
 	SHA3_384(1 << 24),
 	/** SHA3-512 hash. */
-	SHA3_512(1 << 25);
+	SHA3_512(1 << 25),
 	/** CRC32C checksum. */
 	CRC32C(1 << 26),
 	/** Snefru-128 hash. */
 	SNEFRU128(1 << 27),
 	/** Snefru-256 hash. */
-	SNEFRU256(1 << 28),
+	SNEFRU256(1 << 28);
 
 	/** hash_id for the native API */
 	private int hashId;
