@@ -11,12 +11,13 @@ extern "C" {
 /**
  * An element of a list specifying an output format.
  */
-typedef struct print_item {
-	struct print_item *next;
+typedef struct print_item
+{
+	struct print_item* next;
 	unsigned flags;
 	unsigned hash_id;
 	unsigned width;
-	const char *data;
+	const char* data;
 } print_item;
 
 /**
@@ -24,7 +25,7 @@ typedef struct print_item {
  */
 typedef struct print_hash_info
 {
-	const char *name;
+	const char* name;
 	char short_name[20];
 	char short_char;
 } print_hash_info;
@@ -40,13 +41,13 @@ void init_hash_info_table(void);
 void init_printf_format(struct strbuf_t* out);
 
 /* formatted output of hash sums and file information */
-print_item* parse_print_string(const char* format, unsigned *sum_mask);
-void print_line(FILE* out, print_item* list, struct file_info *info);
+print_item* parse_print_string(const char* format, unsigned* sum_mask);
+int print_line(FILE* out, unsigned out_mode, print_item* list, struct file_info* info);
 void free_print_list(print_item* list);
 
 /* SFV format functions */
-void print_sfv_banner(FILE* out);
-int print_sfv_header_line(FILE* out, struct file_t* file, const char* printpath);
+int print_sfv_banner(FILE* out);
+int print_sfv_header_line(FILE* out, unsigned out_mode, struct file_t* file);
 
 #ifdef __cplusplus
 } /* extern "C" */
