@@ -2,6 +2,7 @@
 #ifndef PARSE_CMD_LINE_H
 #define PARSE_CMD_LINE_H
 
+#include <stddef.h>
 #include <stdint.h>
 
 #ifdef __cplusplus
@@ -45,19 +46,22 @@ enum {
 	OPT_LOWERCASE  = 0x8000,
 	OPT_GOST_REVERSE = 0x10000,
 	OPT_BENCH_RAW  = 0x20000,
+	OPT_HEX        = 0x040000,
+	OPT_BASE32     = 0x080000,
+	OPT_BASE64     = 0x100000,
+	OPT_FMT_MODIFIERS = OPT_HEX | OPT_BASE32 | OPT_BASE64,
 
 #ifdef _WIN32
 	OPT_UTF8 = 0x10000000,
-	OPT_ANSI = 0x20000000,
-	OPT_OEM  = 0x40000000,
-	OPT_ENCODING = OPT_UTF8|OPT_ANSI|OPT_OEM,
+	OPT_ENC_WIN = 0x20000000,
+	OPT_ENC_DOS = 0x40000000,
+	OPT_ENCODING = OPT_UTF8 | OPT_ENC_WIN | OPT_ENC_DOS,
 #endif
 
 	FMT_BSD     = 1,
 	FMT_SFV     = 2,
 	FMT_SIMPLE  = 4,
 	FMT_MAGNET  = 8,
-	OPT_FORMAT_MASK = FMT_BSD|FMT_SFV|FMT_SIMPLE|FMT_MAGNET
 };
 
 struct vector_t;
@@ -71,30 +75,31 @@ struct options_t
 	unsigned sum_flags;  /* flags to specify what sums will be calculated */
 	unsigned fmt;        /* flags to specify output format to use */
 	unsigned mode;       /* flags to specify program mode */
-	unsigned openssl_mask;  /* bit-mask for enabled OpenSSL hash functions */
-	const char* config_file; /* config file path */
-	char* printf_str;        /* printf-like format */
+	unsigned openssl_mask;    /* bit-mask for enabled OpenSSL hash functions */
+	const char* config_file;  /* config file path */
+	char* printf_str;         /* printf-like format */
 	opt_tchar* template_file; /* printf-like template file path */
-	opt_tchar* output;       /* file to output calculation or checking results to */
-	opt_tchar* log;          /* file to log percents and other info to */
+	opt_tchar* output;        /* file to output calculation or checking results to */
+	opt_tchar* log;           /* file to log percents and other info to */
+	opt_tchar* update_file;   /* hash file to update */
 	char* embed_crc_delimiter;
 	char  path_separator;
 	int   find_max_depth;
-	struct vector_t *files_accept; /* suffixes of files to process */
-	struct vector_t *files_exclude; /* suffixes of files to exclude from processing */
-	struct vector_t *crc_accept;   /* suffixes of crc files to verify or update */
-	struct vector_t * bt_announce; /* BitTorrent announce URL */
+	struct vector_t* files_accept; /* suffixes of files to process */
+	struct vector_t* files_exclude; /* suffixes of files to exclude from processing */
+	struct vector_t* crc_accept;   /* suffixes of hash files to verify or update */
+	struct vector_t* bt_announce; /* BitTorrent announce URL */
 	size_t bt_piece_length; /* BitTorrent piece length */
 	opt_tchar*  bt_batch_file;   /* path to save a batch torrent to */
 
 	char** argv;
 	int has_files; /* flag: command line contain files */
 	struct file_search_data* search_data; /* files obtained from the command line */
-	struct vector_t *mem; /* allocated memory blocks that must be freed on exit */
+	struct vector_t* mem; /* allocated memory blocks that must be freed on exit */
 };
 extern struct options_t opt;
 
-void read_options(int argc, char *argv[]);
+void read_options(int argc, char* argv[]);
 void options_destroy(struct options_t*);
 
 #ifdef _WIN32
